#include <stdio.h>
#include <string.h>
#include <libgen.h>
#include <sys/stat.h>
#include "dbparser.h"
#include "log.h"

static db_image_list *current_image_list;
static db_image_group *current_image_group;
static db_image *current_image;
static db_photo_album *current_photo_album;

static char db_path[100];

// be is used for compiling on the desktop and determining the system's
// endian setup. but on the ipod it's always little endian, so to make
// it a tiny bit faster, all the functions assume LE on the iPod
#ifndef IPOD
static int be = 1;
#endif

// from make_fw.c
unsigned long switch_32(unsigned long l)
{
#ifndef IPOD
	if (!be) return l;
#endif
	return ((l & 0xff) << 24)
	    | ((l & 0xff00) << 8)
	    | ((l & 0xff0000) >> 8)
	    | ((l & 0xff000000) >> 24);
}

// from make_fw.c
unsigned short switch_16(unsigned short s)
{
#ifndef IPOD
	if (!be) return s;
#endif
	return ((s & 0xff) << 8) | ((s & 0xff00) >> 8);
}

int read_string(FILE *file, int len, char *buffer)
{
	if (fread(buffer, 1, len, file) != len)
		return 0;
		
	buffer[len+1] = 0;
	
	return 1;
}

unsigned long read_longp(FILE *file, int position)
{
	unsigned long l;
	
	if (fseek(file, position, SEEK_CUR) != 0)
		return 0;
	
	fread(&l, 4, 1, file);
	
#ifdef IPOD
	return l;
#else
	return switch_32(l);
#endif
	
	//fread(v, 1, 4, file);
	//return switch_32 ((long)((v[0] << 24) + (v[1] << 16) + (v[2] << 8) + v[3]));
}

unsigned short read_shortp(FILE *file, int position)
{
	unsigned short s;
	
	if (fseek(file, position, SEEK_CUR) != 0)
		return 0;
	
	fread(&s, 2, 1, file);


#ifdef IPOD
	return s;
#else
	return switch_16(s);
#endif
	
	//return switch_16 ((short)((v[0] << 8) + v[1]));
}

unsigned long read_long(FILE *file)
{
	return read_longp(file, 0);
}

unsigned short read_short(FILE *file)
{
	return read_shortp(file, 0);
}


void process_db(FILE *file, int position, void *data)
{
	char header[5] = {0,0,0,0,0};
	unsigned long headerLen = 0;
	int nextPos = -1;
	void *userData = data;
	
	//printf("pos: %X\n", position);
	if (fseek(file, position, SEEK_SET) != 0)
		return;
		
	read_string(file, 4, header);
	headerLen = read_long(file);
	
	//log_message(header);

	//printf("header: %s (len = %X)\n", header, headerLen);
	
	if (strcmp(header, "mhfd") == 0)
	{ 
	}
	else if (strcmp(header, "mhsd") == 0)
	{
		long type = read_longp(file, 4);
		if (type == 3)
			return;
	}
	else if (strcmp(header, "mhli") == 0)
	{
		long numImages = read_long(file);
		
		//printf("db numImages: %l\n", numImages);
		
		// allocate full list of groups here
		current_image_list->groups = (db_image_group *)calloc(numImages, sizeof(db_image_group));		
	}
	else if (strcmp(header, "mhii") == 0)	// each group (list) of images
	{
		long totalLen = read_long(file);
		long numChildren = read_long(file);
		long imageID = read_long(file);
		
		// this fixes a WEIRD "bug" with my db. the first mhii in my db
		// has a header that's 4 bytes longer, and where the number of children
		// value should be, is "mhni". not sure why that's there, but the
		// children value is next, like normal. so i just check for that value
		// here and read the next numbers
		if (numChildren == 0x696E686D)
		{
			numChildren = read_long(file);
			headerLen += 4;
		}
		
		// create new group
		current_image_list->count++;
		current_image_group = &(current_image_list->groups[(current_image_list->count-1)]);
		
		// create new images array
		current_image_list->groups[current_image_list->count-1].images = (db_image *)calloc(numChildren, sizeof(db_image));
		current_image_list->groups[current_image_list->count-1].count = 0;
		
		current_image_group->image_id = imageID;
		//printf("imageID: %X\n", imageID);
	}
	else if (strcmp(header, "mhod") == 0)
	{
		long totalLen = read_long(file);
		short type = read_short(file); // reads 2 bytes for type;

		if (type == 1 || type == 3)
		{
			fgetc(file); // read 1 blank byte;
			short nullChars = fgetc(file); // null char is 1 byte
			long strLen = read_longp(file, 8); // skip 8 bytes of 0s
		
			fseek(file, position + (totalLen - strLen - nullChars), SEEK_SET);
			
			// type 1 = UTF8, type 3 = UTF-16
			//printf("pos: %X\n", ftell(file));
			char text[strLen+1];
			int i, k=0;
			char c;
			for (i=0; i<strLen; i++)
			{
				c = fgetc(file);
				if (c != 0)
					text[k++] = c;
			}
			text[k] = 0;
			//printf("text: %s (%d)\n", text, strlen(text));
			
			if (userData != NULL && (int)userData == 1)
			{
				// reading a name of an album
				current_photo_album->title = strdup(text);
				//printf("text: %s\n", text);
			}
			else
			{
				if (type == 1)
					current_image_group->title = strdup(text);
				else
					current_image->thumbs_path = strdup(text);
			}

			nextPos = ftell(file) + nullChars; // start next read at this location
		}
		else if (type == 2 && (userData != NULL && (int)userData == 1))
		{
			// mhod that has info about the transition effect thingy... skip it
			userData = NULL;
			nextPos = position + totalLen;
		}
	}
	else if (strcmp(header, "mhni") == 0)		// info for an image
	{
		long totalLen = read_long(file);
		long numChildren = read_long(file);
		long correlationID = read_long(file);
		long offset = read_long(file);
		long sizeBytes = read_long(file);
		/*long padding = read_long(file);
		short imgHeight = read_short(file);
		short imgWidth = read_short(file);*/
		
		
		//printf("db w/h: %d/%d (%X)\n", imgWidth, imgHeight, sizeBytes);
		
		//printf("correlationID: %d (%d x %d)\n", correlationID, imgWidth, imgHeight);
		
		int c = (current_image_list->groups[current_image_list->count-1].count++);
		current_image = &(current_image_list->groups[current_image_list->count-1].images[c]);
		current_image->offset = offset;
		current_image->length = sizeBytes;
		current_image->type = 0;
	}
	else if (strcmp(header, "mhla") == 0)	// list of albums
	{
		long numChildren = read_long(file);
		
		current_image_list->albums = (db_photo_album *)calloc(numChildren, sizeof(db_photo_album));
	}
	else if (strcmp(header, "mhba") == 0)	// album
	{
		long totalLen = read_long(file);
		long dataObjects = read_long(file);
		long albumItems = read_long(file);
		long albumType = 0;
		
		fseek(file, position + 30, SEEK_SET);
		albumType = fgetc(file);
		//printf("mhba objects: %d\n", numChildren);
		//printf("album items: %d\n", albumItems);
		
		int c = ++current_image_list->album_count;
		current_image_list->albums[c-1].image_ids = (int *)calloc(albumItems, sizeof(int));
		current_photo_album = &(current_image_list->albums[c-1]);
		current_photo_album->type = albumType;
		
		userData = (void *)1;
	}
	else if (strcmp(header, "mhia") == 0)	// album item
	{
		long imageID = read_longp(file, 8);
		int c = current_photo_album->count++;
		current_photo_album->image_ids[c] = imageID;
	}
	else
	{
		printf("Unused header: %s\n", header);
		return;
	}
	
	if (nextPos == -1)
		nextPos = position + headerLen;
	process_db(file, nextPos, userData);
}

int load_db(const char *fname, db_image_list **image_list)
{
	FILE *file;
	struct stat r;

	//printf("path: %s\n", fname);

	if (stat(fname, &r) != 0)
		return 0;

	strcpy(db_path, fname);
	
	file = fopen(db_path, "rb");
	if (file == NULL)
	{
		log_message("db file null\n");
		return 0;
	}
	
#ifndef IPOD
	// determins endian..
	char ch[4] = { '\0', '\1', '\2', '\3' };
    unsigned i = 0x00010203;
    be = (*((int *)ch) == i);
#endif


	current_image_list = (db_image_list *)calloc(1, sizeof(db_image_list));
	
	process_db(file, 0, NULL);
	fclose(file);
	
	*image_list = current_image_list;

	return 1;
}

void set_size(db_image *img)
{
	// http://ipodlinux.org/ITunesDB#Image_Name
	//printf("size: %d\n", img->length);
	switch (img->length)
	{
		case 691200:			// PhotoPod and VideoPod NTSC image
			img->h = 480;
			img->w = 720;
			img->type = 2;
			break;
		case 153600:			// VideoPod full screen
			img->h = 240;
			img->w = 320;
			img->type = 0;
			img->rotated = 1;
			break;
		case 80000:				// VideoPod album art big version
			img->h = 200;
			img->w = 200;
			break;
		case 77440:				// PhotoPod full screen
			img->h = 176;
			img->w = 220;
			img->type = 1;
			break;
		case 46464:				// Nano full screen
			img->h = 132;
			img->w = 176;
			img->type = 1;
			img->rotated = 1;
			break;
		case 39200:				// PhotoPod album art big version
			img->h = 140;
			img->w = 140;
			break;
		case 22880:				// PhotoPod and VideoPod video preview
			img->h = 88;
			img->w = 130;
			img->rotated = 1;
			break;
		case 20000:				// VideoPod album art small version, Nano album art big version
			img->h = 100;
			img->w = 100;
			img;
		case 6272:				// PhotoPod album art small version
			img->h = 56;
			img->w = 56;
			break;
		case 4100:				// VideoPod list thumbnail
			img->h = 41;
			img->w = 50;
			img->rotated = 1;
			break;
		case 3528:				// Nano album art small version
			img->h = 42;
			img->w = 42;
			break;
		case 3108:				// Nano list thumbnail
			img->h = 37;
			img->w = 42;
		case 2520:				// /PhotoPod list thumbnail
			img->h = 30;
			img->w = 42;
			break;
		default:
			printf("Unknown image size: %d\n", img->length);
			break;
	}
}

void set_images_sizes(db_image_group *group)
{
	int i;
	for (i=0; i<group->count; i++)
		set_size(&(group->images[i]));
}

void RGB565toRGB(unsigned short rgb565, unsigned char *r, unsigned char *g, unsigned char *b, int swap)
{
	unsigned short temp = (swap ? switch_16(rgb565) : rgb565);
	unsigned char red, green, blue;
	red = (temp & 0xF800) >> 11;
	green = (temp & 0x07E0) >> 5;
	blue = (temp & 0x001F);	
	red *= 8;
	green *= 4;
	blue *= 8;
	*r = red;
	*g = green;
	*b = blue;
}

int get_bytes(db_image *image, char **rgb)
{
	int i, l = image->w*image->h*3, k;
	char dir[strlen(db_path)+1];
	dir[0] = 0;
		
	*rgb = (char *)calloc(l, sizeof(char));

	log_message("Start: ");
	log_message(db_path);
	log_message("dir before: ");
	log_message(dir);

	strcpy(dir, db_path);
	//printf("dir: %s\n", dir);

	// strip every up to the first slash.. path must not end with /
	for (i=strlen(dir); i>=0; i--)
	{
		char c = dir[i];
		dir[i] = 0;
		if (c == '/')
			break;
	}
	
	//printf("dir: %s\n", dir);
	log_message("dir after: ");
	log_message(dir);
	
	char full_path[strlen(dir) + strlen(image->thumbs_path)+1];
	full_path[0] = 0;
	strcat(full_path, dir);
	strcat(full_path, image->thumbs_path);

	// replace : with /
	for (i=0; i<strlen(full_path); i++)
		if (full_path[i] == ':')
			full_path[i] = '/';
	
	log_message(full_path);
	//printf("path: %s\n", full_path);
	
	FILE *file = fopen(full_path, "rb");
	if (file == NULL)
	{
		printf("Couldn't open file\n");
		log_message("couldn't open thumbs file");
		return 0;
	}
	
	unsigned char rgb565[image->length];
	fseek(file, image->offset, SEEK_SET);
	if (fread(rgb565, 1, image->length, file) != image->length)
	{
		printf("Couldn't read data\n");
		return 0;
	}
	
	if (image->length % 2 != 0)
	{
		printf("Bad image length: %d\n", image->length);
		return 0;
	}
	
	//printf("image->type: %d\n", image->type);
	
	k = 0;
	for (i=0; i<image->length; i+=2) // we want to read 2 chars at a time (RGB565 = 2 bytes)
	{
		short value = (rgb565[i] << 8) | rgb565[i+1];
		unsigned char r, g, b;
		RGB565toRGB(value, &r, &g, &b, (image->type == 0));
		(*rgb)[k++] = r;
		(*rgb)[k++] = g;
		(*rgb)[k++] = b;
	}

	log_message("Done\n");
		
	// k should = 3*w*h
	return k;
}

int image_to_ppm(db_image *image, const char *fname)
{
	int len;
	char header[20];
	char *bytes;

	sprintf(header, "P6\n%d %d\n255\n", image->w, image->h);

	len = get_bytes(image, &bytes);
	
	FILE *outFile = fopen(fname, "w+");
	if (!outFile)
	{
		printf("Can't open file to write to...\n");
		return 0;
	}
	fwrite(header, 1, strlen(header), outFile);
	fwrite(bytes, 1, len, outFile);
	fclose(outFile);
	
	free(bytes);
	
	return 1;
}

void free_image_list(db_image_list *list)
{
	if (list == NULL)
		return;
	
	int i, j;
	for (i=0; i<list->count; i++)
	{
		for (j=0; j<list->groups[i].count; j++)
			free(list->groups[i].images[j].thumbs_path);
		
		free(list->groups[i].title);
		free(list->groups[i].images);
	}

	for (i=0; i<list->album_count; i++)
	{
		free(list->albums[i].title);
		free(list->albums[i].image_ids);
	}
	
	free(list->groups);
	free(list->albums);
	free(list);
}
