/* Free for absolutely any use whatsoever. */

#include <stdio.h>
#include <string.h>
#include "pz.h"
#include "dbparser.h"
#include "log.h"


#ifdef IPOD
static char *photosPath[] = { "/Photos/Photo Database", "/hp/Photos/Photo Database", NULL };
static char *artworkPath[] = { "/iPod_Control/Artwork/ArtworkDB", "/hp/iPod_Control/Artwork/ArtworkDB", NULL };
#else
// "/Users/kainjow/Documents/Source Code/pz2/podzilla2/modules/photos/Photo Database/Photo Database"
static char *photosPath[] = { "/Users/kainjow/Documents/Source Code/photosdb/Photos/Photo Database", "/bob/smith", NULL };
static char *artworkPath[] = { "/Users/kainjow/Documents/Source Code/photosdb/Artwork/ArtworkDB", NULL };
#endif


static PzModule *module = NULL;
static TWindow *images_window = NULL;
static db_image_list *image_list = NULL;

static db_photo_album *album = NULL;
static int img_index;

static unsigned short *framebuffer = NULL;
static int drawing_mode = 0;


static int sort_images(const void *a, const void *b)
{
	db_image *va = (db_image *)a;
	db_image *vb = (db_image *)b;
	
	if (va->length > vb->length)
		return 1;
	else if (va->length < vb->length);
		return -1;
	return 0;
}

int image_for_screen(db_image_group *group)
{
	// returns the index of the image with the 
	int b = ttk_screen->w * ttk_screen->h * 2;
	int i;
	
	/*if (!group->sorted)
	{
		qsort(group->images, group->count, sizeof(db_image), sort_images);
		group->sorted = 1;
	}*/
	qsort(group->images, group->count, sizeof(db_image), sort_images);

	for (i=group->count-1; i>=0; i--)
	{
		db_image *img = &(group->images[i]);
		//printf("img->length: %d\n", img->length);
		if (img->length > b) // skip images that are bigger than the iPod's screen
			continue;
		return i;
	}
	
	return 0;
}

void draw(struct TWidget *this, ttk_surface srf)
{
	if (album == NULL)
		return;
		
	db_image_group *group = NULL;
	db_image *img = NULL;
	char *rgb_bytes = NULL;
	int w, h, k=0, j,x,y,cw,xs,ys,xe,ye, r,g,b,a,p;
	
	// find the group..
	int imgID = album->image_ids[0], i;
	for (i=0; i<image_list->count; i++)
	{
		if (image_list->groups[i].image_id == album->image_ids[img_index])
		{
			group = &(image_list->groups[i]);
			break;
		}
	}
		
	if (group == NULL || group->count == 0)
	{
		printf("null group\n");
		return;
	}
	
	// sets the w/h of the images in the group. must be called before you
	// access the image's w/h values
	set_images_sizes(group);
	
	img = &(group->images[image_for_screen(group)]);

	k = get_bytes(img, &rgb_bytes);
	
	//ttk_surface_get_dimen(srf, &w, &h);
	//ttk_fillrect(srf, 0, 0, w, h, /*ttk_ap_getx("window.bg")->color*/ttk_makecol(BLACK));

	//printf("length: %d\n", img->length);
	//printf("rotated: %d\n", img->rotated);
	
	if (img->rotated)
	{
		xs = (ttk_screen->h - img->h) / 2;
		ys = (ttk_screen->w - img->w) / 2;
		xe = img->h;
		ye = img->w;
	}
	else
	{
		xs = (ttk_screen->w - img->w) / 2;
		ys = (ttk_screen->h - img->h) / 2;
		xe = img->w;
		ye = img->h;
	}
	
	xe += xs;
	ye += ys;
	
	j = 0;
	for (x=xs; x<xe; x++) 
	{
		for (y=ys; y<ye; y++)
		{
			r = rgb_bytes[j];
			g = rgb_bytes[j+1];
			b = rgb_bytes[j+2];
			
			switch (drawing_mode)
			{
				case 0: // normal
					break;
				case 1: // grayscale
					a = (r > g ? r : g);
					a = (a > b ? a : b);
					//a = 0.299*r + 0.587*g + 0.114*b; //(r + g + b) / 3;
					r = a; g = a; b = a;
					break;
				case 2: // invert
					r = 255-r; g = 255-g; b = 255-b;
					break;
				case 3: // noise
					r += rand()%((255-r)/5); g += rand()%((255-g)/5); b += rand()%((255-b)/5);
					break;
			}
			
			if (img->rotated)
				p = x*ttk_screen->w + y;
			else
				p = y*ttk_screen->w + x;

			framebuffer[p] = ttk_makecol(r, g, b);
			j += 3;
		}
	}
	memcpy(srf->pixels, framebuffer, ttk_screen->w * ttk_screen->h * ttk_screen->srf->format->BytesPerPixel);

	free(rgb_bytes);
}

void prev_image(TWindow *win)
{
	if (img_index > 0)
		img_index--;
		
	win->dirty = 1;
}

void next_image(TWindow *win)
{
	if (img_index < album->count-1)
		img_index++;
	win->dirty = 1;
}

int images_down (TWidget *this, int button)
{
	if (button == TTK_BUTTON_MENU)
	{
		pz_hide_window(this->win);
		return 0;
	}
	else if (button == TTK_BUTTON_NEXT)
	{
		next_image(this->win);
		return 0;
    }
    else if (button == TTK_BUTTON_PREVIOUS)
    {
		prev_image(this->win);
		return 0;
    }
    else if (button == TTK_BUTTON_ACTION)
    {
    	drawing_mode++;
    	if (drawing_mode == 4)
    		drawing_mode = 0;
    	this->dirty = 1;
    }
    
    return TTK_EV_UNUSED;
}

int images_scroll (TWidget *this, int dir)
{
	return 0;
}

void show_album(db_photo_album *a)
{
	static TWidget *images_widget = NULL;

	if (a->count == 0)
		return;

	if (images_window == NULL)
	{
		images_window = ttk_new_window();
	}
	else
	{
		ttk_remove_widget(images_window, images_widget);
		ttk_free_widget(images_widget);
	}
	
	img_index = 0;
	
	images_widget = ttk_new_widget(ttk_screen->w - ttk_screen->wx, ttk_screen->h - ttk_screen->wy);
	images_widget->down = images_down;
	images_widget->draw = draw;
	images_widget->scroll = images_scroll;
    images_widget->dirty = 1;
    images_widget->focusable = 1;
    
    images_window->dirty = 1;

	ttk_window_hide_header(images_window);
	ttk_add_widget(images_window, images_widget);
	ttk_show_window(images_window);
}

int load()
{
	int i = 0;
	while (photosPath[i] && !load_db(photosPath[i], &image_list))
		i++;
		
	return (image_list == NULL ? 0 : image_list->count);
}

int menu_down(TWidget *this, int button)
{
	if (button == TTK_BUTTON_ACTION)
	{
		if (ttk_menu_get_selected_item(this) == NULL)
		{
			ttk_hide_window(this->win);
		}
		else
		{
			album = (db_photo_album *)(ttk_menu_get_selected_item(this)->data);
			show_album(album);
		}
	}

    return ttk_menu_down(this, button);
}


TWindow *setup()
{
	TWindow *win = NULL;
    TWidget *menu = NULL;
    int i, j;
    
	load();
	menu = ttk_new_menu_widget(0, ttk_menufont, ttk_screen->w - ttk_screen->wx, ttk_screen->h - ttk_screen->wy);
	menu->down = menu_down;
	
    if (image_list != NULL)
	{
		for (i=0; i<image_list->album_count; i++)
		{
			db_photo_album *album = &(image_list->albums[i]);
			
			ttk_menu_item *item = calloc (1, sizeof(struct ttk_menu_item));
			
			if (album->type == 1)
				item->name = strdup("Photo Library");
			else
				item->name = strdup(album->title);
			item->data = (void *)album;
	
			ttk_menu_append(menu, item);
		}

	   	framebuffer = malloc(ttk_screen->w * ttk_screen->h * ttk_screen->srf->format->BytesPerPixel);
	}
    
    win = ttk_new_window();
    ttk_add_widget(win, menu);
    ttk_window_set_title(win, "Photos");

    return win;
}

void cleanup()
{
	free(framebuffer);
	free_image_list(image_list);
}

void init() 
{
	ttk_menu_item *photosMenu;
    module = pz_register_module("photos", cleanup);

	photosMenu = pz_menu_add_action("/Extras/Utilities/Photos", setup);
	
	srand(time(NULL));
}

PZ_MOD_INIT (init)
